// This program doesn't support Unicode.
// #define UNICODE

// for Win98 or later.
#define _WIN32_WINDOWS 0x0410

#include <windows.h>
#include "runtime.h"
#include "resource.h"
#include "bit.h"
#include "main.h"
#include "smaf.h"
#include "detail.h"
#include "emusmw5.h"


//public
UINT DrawFunc = 2;
UINT PlayingPos = 0;

//private
static UINT RollHeight = 4;
static UINT RollWidth = 25;
static UINT FollowPlay = 0;
static int FollowSpace = -1000;
static int VisibleTrack = -1;
static BOOL PlayVisibleOnly = FALSE;

//?
HWND hMainWnd;
HWND hDetailWnd;

//static VOICE voices[256];
static VOICE* voices = NULL;
static UINT voicesmax = 0;

COLORREF trackcolors[16] = {
	RGB(192, 192, 192),
	RGB(255, 128, 128),
	RGB(128, 255, 128),
	RGB(128, 128, 255),
	RGB(255, 255, 128),
	RGB(128, 255, 255),
	RGB(255, 128, 255),
	RGB(255, 255, 255),
	RGB(  0,   0,   0),
	RGB(128,   0,   0),
	RGB(  0, 128,   0),
	RGB(  0,   0, 128),
	RGB(128, 128,   0),
	RGB(  0, 128, 128),
	RGB(128,   0, 128),
	RGB(128, 128, 128),
};

BOOL CALLBACK MessageDlgWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	if(msg == WM_INITDIALOG)
		return TRUE;
	if(msg == WM_COMMAND && (wParam == IDOK || wParam == IDCANCEL))
		EndDialog(hWnd, wParam);

	return FALSE;
}

void draw(HWND hWnd, HDC hdc, int width, int height, SMAF* smaf, DWORD playingpos)
{
	HFONT hfold;
	int scrx, scry;

	SetTextColor(hdc, RGB(0, 0, 0));
	SetBkMode(hdc, TRANSPARENT);

	FillRectangle(hdc, 0, 0, width, height, RGB(255, 255, 255));

	{
		LOGFONT lf = {0};

		lf.lfCharSet = DEFAULT_CHARSET;
		strcopy(lf.lfFaceName, "lr SVbN");
		lf.lfHeight = -12;

		hfold = SelectObject(hdc, CreateFontIndirect(&lf));
	}

	{
		SCROLLINFO vsi, hsi;

		vsi.cbSize = hsi.cbSize = sizeof(SCROLLINFO);
		vsi.fMask = hsi.fMask = SIF_POS;
		GetScrollInfo(hWnd, SB_VERT, &vsi);
		GetScrollInfo(hWnd, SB_HORZ, &hsi);

		scrx = hsi.nPos;
		scry = vsi.nPos;
	}

	// draw pianoroll
	{
		int i;
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(220, 220, 220)));
		COLORREF oldcolor = SetTextColor(hdc, RGB(192, 192, 192));
		EVENT* e;

		// draw horizontal grid
		for(i = 0; i < 128; i++)
		{
			MoveToEx(hdc, 0, (127 - i) * RollHeight - scry, NULL);
			LineTo(hdc, width, (127 - i) * RollHeight - scry);

			if(!(i % 12))
			{
				FillRectangle(hdc, 0, (127 - i) * RollHeight - scry + 1, width, (127 - i + 1) * RollHeight - scry, RGB(240, 240, 240));
			}
		}

		// draw vertical grid
		for(i = -(scrx * (int)RollWidth % 100); i < width * (int)RollWidth; i += 100)
		{
			HPEN hpold;

			if(!((i + scrx * RollWidth) % (50 * RollWidth)))
				hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(192, 192, 192)));

			MoveToEx(hdc, i / RollWidth, 0, NULL);
			LineTo(hdc, i / RollWidth, height);

			if(!((i + scrx * RollWidth) % (50 * RollWidth)))
				DeleteObject(SelectObject(hdc, hpold));
		}

		if(smaf)
			for(e = smaf->events; e; e = e->next)
			{
				DWORD gtime;

				if((e->status & 0xF0) != 0x80 && (e->status & 0xF0) != 0x90)	// Not NoteON
					continue;

				//if(scrx + width < (signed)e->time)	// Over. no draw needed any more.
				//	break;

				if(VisibleTrack != -1 && (e->status & 0x0F) != VisibleTrack)	// NoVisible
					continue;

				if((e->status & 0xF0) == 0x80)	// NonVel
					gtime = GetVVal(e->data + 1) * smaf->gbase;
				else
					gtime = GetVVal(e->data + 2) * smaf->gbase;

				FillRectangle(hdc,	e->time / RollWidth - scrx,
									(127 - e->data[0]) * RollHeight - scry,
									(e->time + gtime) / RollWidth - scrx,
									((127 - e->data[0]) + 1) * RollHeight - scry,
									RGB(128, 64, 128));
				FillRectangle(hdc,	e->time / RollWidth - scrx + 1,
									(127 - e->data[0]) * RollHeight - scry + 1,
									(e->time + gtime) / RollWidth - scrx - 1,
									((127 - e->data[0]) + 1) * RollHeight - scry - 1,
									trackcolors[e->status & 0x0F]);
			}

		// draw vertical grid text
		for(i = -(scrx * (int)RollWidth % 100); i < width * (int)RollWidth; i += 100)
		{
			if((i + scrx * RollWidth) % (50 * RollWidth))
				continue;
			MoveToEx(hdc, i / RollWidth + 2, 0, NULL);
			PutText(hdc, "%ums", i + scrx * RollWidth);
		}

		// draw horizontal grid text
		for(i = 0; i < 128; i++)
		{
			LPCTSTR oto[] = {TEXT("C"), TEXT("C#"), TEXT("D"), TEXT("D#"), TEXT("E"), TEXT("F"), TEXT("F#"), TEXT("G"), TEXT("G#"), TEXT("A"), TEXT("A#"), TEXT("B")};
			MoveToEx(hdc, 1, (127 - i) * RollHeight - scry + 1, NULL);
			PutText(hdc, "%-2s %d=%d", oto[i % 12], i / 12 - 2, i);
		}

		SetTextColor(hdc, oldcolor);
		DeleteObject(SelectObject(hdc, hpold));
	}

	// Start
	if(smaf && smaf->start)
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(0, 128, 0)));

		MoveToEx(hdc, (smaf->start->time - smaf->start->duration) / RollWidth - scrx, 0, NULL);
		LineTo(hdc, (smaf->start->time - smaf->start->duration) / RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}
	// End
	if(smaf && smaf->stop)
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(128, 0, 128)));

		MoveToEx(hdc, smaf->stop->time / RollWidth - scrx, 0, NULL);
		LineTo(hdc, smaf->stop->time / RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}

	// Current
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(255, 0, 0)));

		MoveToEx(hdc, playingpos / RollWidth - scrx, 0, NULL);
		LineTo(hdc, playingpos / RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}

	if(VisibleTrack != -1)
	{
		SIZE s;
		TCHAR ts[32];

		SetBkMode(hdc, OPAQUE);
		SetBkColor(hdc, RGB(255, 255, 255));
		SetTextColor(hdc, RGB(255, 0, 0));

		wsprintf(ts, "Viewing Track %d", VisibleTrack + 1);

		GetTextExtentPoint32(hdc, ts, lstrlen(ts), &s);
		TextOut(hdc, width - s.cx, height - s.cy, ts, lstrlen(ts));
	}

	DeleteObject(SelectObject(hdc, hfold));

	return;
}

void getTrackMenuItem(int id, LPTSTR str, COLORREF* fg, COLORREF* bg)
{
	COLORREF dummy;

	if(!fg)
		fg = &dummy;
	if(!bg)
		bg = &dummy;

	switch(id)
	{
	case ID_VTRACK1:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[0];	lstrcpy(str, "1:Track1(Silver)");		break;
	case ID_VTRACK2:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[1];	lstrcpy(str, "2:Track2(Red)");			break;
	case ID_VTRACK3:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[2];	lstrcpy(str, "3:Track3(Green)");		break;
	case ID_VTRACK4:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[3];	lstrcpy(str, "4:Track4(Blue)");			break;
	case ID_VTRACK5:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[4];	lstrcpy(str, "5:Track5(Yellow)");		break;
	case ID_VTRACK6:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[5];	lstrcpy(str, "6:Track6(Cyan)");			break;
	case ID_VTRACK7:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[6];	lstrcpy(str, "7:Track7(Purple)");		break;
	case ID_VTRACK8:	*fg = RGB(  0,   0,   0);	*bg = trackcolors[7];	lstrcpy(str, "8:Track8(White)");		break;
	case ID_VTRACK9:	*fg = RGB(255, 255, 255);	*bg = trackcolors[8];	lstrcpy(str, "Q:Track9(Black)");		break;
	case ID_VTRACK10:	*fg = RGB(255, 255, 255);	*bg = trackcolors[9];	lstrcpy(str, "W:Track10(DarkRed)");		break;
	case ID_VTRACK11:	*fg = RGB(255, 255, 255);	*bg = trackcolors[10];	lstrcpy(str, "E:Track11(DarkGreen)");	break;
	case ID_VTRACK12:	*fg = RGB(255, 255, 255);	*bg = trackcolors[11];	lstrcpy(str, "R:Track12(DarkBlue)");	break;
	case ID_VTRACK13:	*fg = RGB(255, 255, 255);	*bg = trackcolors[12];	lstrcpy(str, "T:Track13(DarkYellow)");	break;
	case ID_VTRACK14:	*fg = RGB(255, 255, 255);	*bg = trackcolors[13];	lstrcpy(str, "Y:Track14(DarkCyan)");	break;
	case ID_VTRACK15:	*fg = RGB(255, 255, 255);	*bg = trackcolors[14];	lstrcpy(str, "U:Track15(DarkPurple)");	break;
	case ID_VTRACK16:	*fg = RGB(255, 255, 255);	*bg = trackcolors[15];	lstrcpy(str, "I:Track16(Gray)");		break;
	}

	return;
}

void seek(SMAF* smaf, UINT ppos, BOOL emuseek, BOOL forcemove)
{
	PlayingPos = ppos;

	if(emuseek)
	{
		if(!EmuSeek(PlayingPos - (smaf->start->time - smaf->start->duration)))
		{
			MessageBox(hMainWnd, "Seek failed", "Error", MB_ICONERROR | MB_OK);
		}
	}

	if(forcemove)
	{
		if(FollowPlay)
			SetScrollPos(hMainWnd, SB_HORZ, (int)(PlayingPos + FollowSpace) / (int)RollWidth, TRUE);
	}else
	{
		if(FollowPlay == 1 || (FollowPlay == 2 && GetAsyncKeyState(VK_SHIFT) & 0x8000))
			SetScrollPos(hMainWnd, SB_HORZ, (int)(PlayingPos + FollowSpace) / (int)RollWidth, TRUE);
	}

	InvalidateRect(hMainWnd, NULL, FALSE);
	InvalidateRect(hDetailWnd, NULL, FALSE);

	return;
}

// check this function
void setExclusiveFMCh(BYTE* data, CHPARAM* chp)
{
	data[0] = 0xF0;
	data[1] = 0x43;
	data[2] = 0x79;
	data[3] = 0x07;
	data[4] = 0x7F;
	data[5] = 0x01;
	data[6] = chp->bm;
	data[7] = chp->bl;
	data[8] = chp->pc;
	data[9] = chp->mapnote;
	data[12] = chp->drumnote;
	data[11] &= ~0x30;
	data[11] |= ((chp->lfo & 0x02) ? 0x10 : 0) | ((chp->pan & 0x10) ? 0x20 : 0);
	data[13] &= ~0x78;
	data[13] |= ((chp->pan & 0x0F) << 3);
	data[14] |= ((chp->lfo & 0x01) ? 0x40 : 0) | ((chp->pe) ? 0x20 : 0) | (chp->alg);

	return;
}

// CHECK THIS FUNCTION
void setExclusiveFMOp(BYTE* data, OPPARAM* opp)
{
	// X1 XX XX XX 23 45 67 89 AX BC DE FG

	data[0] &= ~0x0F;
	data[0] |= ((opp->sr & 0x08) ? 8 : 0) | ((opp->rr & 0x08) ? 4 : 0) | ((opp->ar & 0x08) ? 2 : 0) | ((opp->tl & 0x20) ? 1 : 0);

	data[4] = ((opp->sr & 0x07) << 4) | (opp->xof ? 8 : 0) | (opp->sus ? 2 : 0) | (opp->ksr ? 1 : 0);
	data[5] = ((opp->rr & 0x07) << 4) | (opp->dr);
	data[6] = ((opp->ar & 0x07) << 4) | (opp->sl);
	data[7] = ((opp->tl & 0x1F) << 2) | (opp->ksl);
	data[8] &= ~0x30;
	data[8] |= ((opp->multi & 0x08) ? 0x20 : 0) | ((opp->ws & 0x10) ? 0x10 : 0);
	data[9] = ((opp->dam) << 5) | ((opp->eam) ? 0x10 : 0) | ((opp->dvb) << 1) | ((opp->evb) ? 1 : 0);
	data[10] = ((opp->multi & 0x07) << 4) | (opp->dt);
	data[11] = ((opp->ws & 0x0F) << 3);

	return;
}

UINT setExclusiveFMAll(BYTE* data, CHPARAM* chp, OPPARAM* opps)
{
	UINT size;
	UINT i;

	for(i = 0; i < 0x30; i++)
		data[i] = 0;

	setExclusiveFMCh(data, chp);

	setExclusiveFMOp(data + 11, opps + 0);
	setExclusiveFMOp(data + 19, opps + 1);

	if(chp->alg <= 1)
	{
		size = 0x20;
	}else
	{
		setExclusiveFMOp(data + 27, opps + 2);
		setExclusiveFMOp(data + 35, opps + 3);
		size = 0x30;
	}

	data[size - 1] = 0xF7;

	return size;
}

UINT setExclusivePCMAll(BYTE* data, CHPARAM* chp, OPPARAM* opp)
{
	UINT i;

	for(i = 0; i < 0x1F; i++)
		data[i] = 0;

	data[0] = 0xF0;
	data[1] = 0x43;
	data[2] = 0x79;
	data[3] = 0x07;
	data[4] = 0x7F;
	data[5] = 0x01;
	data[6] = chp->bm;
	data[7] = chp->bl;
	data[8] = chp->pc;
	data[9] = chp->mapnote;
	data[10] = 0x01;
	data[11] |= ((chp->pan & 0x10) ? 0x10 : 0) | ((chp->fs & 0x80) ? 0x20 : 0) | ((chp->fs & 0x8000) ? 0x40 : 0);
	data[11] |= ((opp->ar & 0x08) ? 1 : 0) | ((opp->rr & 0x08) ? 2 : 0) | ((opp->sr & 0x08) ? 4 : 0) | ((chp->lfo & 0x02) ? 8 : 0);
	data[12] = (BYTE)((chp->fs & 0x7F00) >> 8);
	data[13] = (BYTE)(chp->fs & 0x7F);
	data[14] = ((chp->pan & 0x0F) << 3) | ((chp->pe) ? 1 : 0);
	data[15] = (chp->lfo & 0x01) ? 0x40 : 0;

	data[16] = ((opp->sr & 0x07) << 4) | ((opp->xof) ? 8 : 0) | ((opp->sus) ? 4 : 0);
	data[17] = ((opp->rr & 0x07) << 4) | (opp->dr);
	data[18] = ((opp->ar & 0x07) << 4) | (opp->sl);
	data[19] |= ((opp->tl & 0x20) ? 0x40 : 0);
	data[19] |= ((chp->lp & 0x80) ? 2 : 0);
	data[20] = ((opp->tl) & 0x1F) << 2;
	data[21] = ((opp->dam) << 5) | ((opp->eam) ? 0x10 : 0) | ((opp->dvb) << 1) | ((opp->evb) ? 1 : 0);
	data[24] = (BYTE)((chp->lp & 0x7F00) >> 8);
	data[25] = (BYTE)(chp->lp & 0x7F);
	data[26] = (BYTE)((chp->ep & 0x7F00) >> 8);
	data[27] = ((chp->rm) ? 0x20 : 0) | ((chp->ep & 0x80) ? 0x40 : 0);
	data[28] = (BYTE)(chp->ep & 0x7F);
	data[29] = chp->wavno;
	data[30] = 0xF7;

	return 0x1F;
}

void LoadVoices(LPCTSTR fn)
{
	BYTE* vm;

	{
		HANDLE hf;
		DWORD read;

		if((hf = CreateFile(fn, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
		{
			MessageBox(NULL, "DefMA3_16.vm3 couldn't open\nIt isn't required, but there are some affections.", "Error", MB_ICONERROR | MB_OK);
			return;
		}

		if(!(vm = halloc(GetFileSize(hf, NULL))))
		{
			CloseHandle(hf);
			MessageBox(NULL, "Couldn't allocate memory", "Error", MB_ICONERROR | MB_OK);
			return;
		}

		if(!ReadFile(hf, vm, GetFileSize(hf, NULL), &read, NULL))
			read = 0;

		if(read != GetFileSize(hf, NULL))
		{
			hfree(vm);
			CloseHandle(hf);
			MessageBox(NULL, "Couldn't read correctly", "Error", MB_ICONERROR | MB_OK);
			return;
		}

		CloseHandle(hf);
	}

	if(COMPARE4(vm, "FMM3"))
	{
		hfree(vm);
		MessageBox(NULL, "This is not FMM3", "Error", MB_ICONERROR | MB_OK);
		return;
	}

	{
		BITFILE* bf;
		int cnt, i;
		BYTE* p;
		DWORD size;

		p = vm + 8;
		size = SwapDword(vm + 4);

		for(cnt = 0; (DWORD)(p - vm - 8) < size/* && cnt < 128*/; cnt++)
		{
			BYTE rsize;

			UINT rec;

			rec = voicesmax;

			if(voicesmax < rec + 1)
			{
				voicesmax = rec + 1;
				voices = hrealloc(voices, (voicesmax) * sizeof(VOICE));
			}

			p += 2;	// skip record no
			//: VVal
			rsize = *p;						p += 1;

			voices[rec].chp.bm = *p;		p += 1;
			voices[rec].chp.bl = *p;		p += 1;
			voices[rec].chp.pc = *p;		p += 1;
			voices[rec].chp.mapnote = *p;	p += 1;

			voices[rec].chp.type = *p;		p += 1;

			memcopy(voices[rec].name, p, 16);
			voices[rec].name[16] = '\0';
			p += 16;

			// Voiceǂݍ݂PCMԂ(L֥`)
			switch(voices[rec].chp.type)
			{
			case 0:
				voices[rec].chp.drumnote = *p;	p += 1;
				rsize -= 1;

				bf = bitopen(p, 55);

				voices[rec].chp.pan = bitnread(bf, 5);
				bitnread(bf, 3);

				voices[rec].chp.lfo = bitnread(bf, 2);
				voices[rec].chp.pe = bitnread(bf, 1);
				bitnread(bf, 2);
				voices[rec].chp.alg = bitnread(bf, 3);

				for(i = 0; i < 4; i++)
				{
					voices[rec].opp[i].sr = bitnread(bf, 4);
					voices[rec].opp[i].xof = bitread(bf);
					bitread(bf);
					voices[rec].opp[i].sus = bitread(bf);
					voices[rec].opp[i].ksr = bitread(bf);

					voices[rec].opp[i].rr = bitnread(bf, 4);
					voices[rec].opp[i].dr = bitnread(bf, 4);

					voices[rec].opp[i].ar = bitnread(bf, 4);
					voices[rec].opp[i].sl = bitnread(bf, 4);

					voices[rec].opp[i].tl = bitnread(bf, 6);
					voices[rec].opp[i].ksl = bitnread(bf, 2);

					voices[rec].opp[i].dam = bitnread(bf, 3);
					voices[rec].opp[i].eam = bitread(bf);
					voices[rec].opp[i].dvb = bitnread(bf, 3);
					voices[rec].opp[i].evb = bitread(bf);

					voices[rec].opp[i].multi = bitnread(bf, 4);
					bitread(bf);
					voices[rec].opp[i].dt = bitnread(bf, 3);

					voices[rec].opp[i].ws = bitnread(bf, 5);
					voices[rec].opp[i].fb = bitnread(bf, 3);
				}

				bitclose(bf);
				break;
			case 1:
				voices[rec].chp.fs = SwapWord(p);	p += 2;
				rsize -= 2;

				bf = bitopen(p, 14);

				voices[rec].chp.pan = bitnread(bf, 5);
				bitnread(bf, 2);
				voices[rec].chp.pe = bitread(bf);

				voices[rec].chp.lfo = bitnread(bf, 2);
				bitnread(bf, 6);

				voices[rec].opp[0].sr = bitnread(bf, 4);
				voices[rec].opp[0].xof = bitread(bf);
				bitread(bf);
				voices[rec].opp[0].sus = bitread(bf);
				bitread(bf);

				voices[rec].opp[0].rr = bitnread(bf, 4);
				voices[rec].opp[0].dr = bitnread(bf, 4);

				voices[rec].opp[0].ar = bitnread(bf, 4);
				voices[rec].opp[0].sl = bitnread(bf, 4);

				voices[rec].opp[0].tl = bitnread(bf, 6);
				bitnread(bf, 2);

				bitread(bf);
				voices[rec].opp[0].dam = bitnread(bf, 2);
				voices[rec].opp[0].eam = bitread(bf);
				bitread(bf);
				voices[rec].opp[0].dvb = bitnread(bf, 2);
				voices[rec].opp[0].evb = bitread(bf);

				bitnread(bf, 16);

				voices[rec].chp.lp = bitnread(bf, 16);
				voices[rec].chp.ep = bitnread(bf, 16);

				voices[rec].chp.rm = bitread(bf);
				bitnread(bf, 4);
				voices[rec].chp.wavno = bitnread(bf, 3);

				bitclose(bf);

				break;
			}

			p += rsize - 5 - 16;
		}
	}

	hfree(vm);

	return;
}

void setVoice(UINT rec, BOOL forcelsb, UINT lsb, BOOL forcepch, UINT pch)
{
	BYTE cmd[0x30] = {0};
	UINT cmdsize;

	CHPARAM chp = voices[rec].chp;
	OPPARAM opp[4];

	opp[0] = voices[rec].opp[0];
	opp[1] = voices[rec].opp[1];
	opp[2] = voices[rec].opp[2];
	opp[3] = voices[rec].opp[3];

	//chp.bm = 124;
	//chp.bl = banklsb;
	if(forcelsb)
		chp.bl = lsb;
	if(forcepch)
		chp.pc = pch;

	if(chp.type == 0x00)
		cmdsize = setExclusiveFMAll(cmd, &chp, opp);
	else
		cmdsize = setExclusivePCMAll(cmd, &chp, opp);

	if(EmuSetMidiMsg(cmd, cmdsize))
	{
		TCHAR ts[256];

		wsprintf(ts, "PresetVoices EmuSetMidiMsg(Assign) failed at Bank%d:%d, PC%d, Note%d", chp.bm, chp.bl, chp.pc, chp.mapnote);
		MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);
		return;
	}

	return;
}

void PresetVoices(void)
{
	UINT i;

	// normal set
	for(i = 0; i < voicesmax; i++)
	{
		int rec = i;
		/*
		//ݒ胊Zbg?Kvۂ?
		{
			BYTE cmd[0x0B] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x02, voices[rec].chp.bm, voices[rec].chp.bl, voices[rec].chp.pc, voices[rec].chp.mapnote, 0xF7};

			if(EmuSetMidiMsg(cmd, 0x0B))
			{
				TCHAR ts[256];

				wsprintf(ts, "PresetVoices EmuSetMidiMsg(Erase) failed at Bank%d:%d, PC%d, Note%d", voices[rec].chp.bm, voices[rec].chp.bl, voices[rec].chp.pc, voices[rec].chp.mapnote);
				MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);
				return;
			}
		}
		*/

		//*
		if(voices[rec].chp.bm == 124 && voices[rec].chp.bl == 0)
		{
			int j;

			for(j = 0; j < 10; j++)
				setVoice(rec, TRUE, j, FALSE, 0);
		}else
		//*/
		if(voices[rec].chp.bm == 125 && voices[rec].chp.bl == 0)
		{
			int j;

			setVoice(rec, FALSE, 0, FALSE, 0);
			for(j = 2; j < 10; j++)
				setVoice(rec, FALSE, 0, TRUE, j);
		}
		else
			setVoice(rec, FALSE, 0, FALSE, 0);
	}

	return;
}

void InitMidiMode(void)
{
	// terminate
	// after
	//Emu526829();

	// MONOPHOLIC EXCLUSIVES?
	/*
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x13, 0x00, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(NULL, "InitMidiMode F0 43 79 06 7F 13 00 F7 failed", "Error", MB_ICONERROR | MB_OK);
			return;
		}
	}
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x11, 0x00, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(NULL, "InitMidiMode F0 43 79 06 7F 11 00 F7 failed", "Error", MB_ICONERROR | MB_OK);
			return;
		}
	}
	//*/

	/*
	Emu5282();

	// Ȃ?
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x13, 0x00, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(hWnd, "SetMidiMsgX1 failed", "Error", MB_ICONERROR | MB_OK);
			break;
		}
	}
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x11, 0x00, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(hWnd, "SetMidiMsgX2 failed", "Error", MB_ICONERROR | MB_OK);
			break;
		}
	}

	// Ȃ?
	// CC BankSelect
	{
		BYTE cmd[] = {0xB0, 0x00, 0x7C};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(hWnd, "SetMidiMsg5 failed", "Error", MB_ICONERROR | MB_OK);
			break;
		}
	}
	{
		BYTE cmd[] = {0xB0, 0x20, 0x00};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(hWnd, "SetMidiMsg6 failed", "Error", MB_ICONERROR | MB_OK);
			break;
		}
	}
	// PC
	{
		BYTE cmd[] = {0xC0, 0};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(hWnd, "SetMidiMsg7 failed", "Error", MB_ICONERROR | MB_OK);
			break;
		}
	}
	*/

	/*
	// unknown exclusive
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x02, 0x7C, 0x00, 0x00, 0x00, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(hWnd, "SetMidiMsgX4 failed", "Error", MB_ICONERROR | MB_OK);
			break;
		}
	}
	*/

	// Reset
	{
		BYTE cmd[] = {0xF0, 0x43, 0x79, 0x06, 0x7F, 0x7F, 0xF7};
		if(EmuSetMidiMsg(cmd, sizeof(cmd)))
		{
			MessageBox(NULL, "SetMidiMsg ini failed", "Error", MB_ICONERROR | MB_OK);
			return;
		}
	}

	EmuMIDISetVolume();

	PresetVoices();

	return;
}

void OpenSMAFCommand(HWND hWnd, SMAF** smaf, LPCTSTR fname)
{
	if(*smaf)
		SendMessage(hWnd, WM_COMMAND, ID_CLOSE, 0);

	PlayingPos = 0;

	*smaf = LoadSmaf(fname);

	SetScrollPos(hWnd, SB_HORZ, 0, TRUE);
	SetScrollPos(hWnd, SB_VERT, 0, TRUE);

	SendMessage(hDetailWnd, WM_USER, 0, (LPARAM)*smaf);

	{
		TCHAR ts[32];

		wsprintf(ts, "MMF Tool %06ums", PlayingPos);
		SetWindowText(hWnd, ts);
	}

	InvalidateRect(hWnd, NULL, FALSE);

	return;
}

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static BOOL emusounding = FALSE;
	static BOOL midisounding = FALSE;
	static SMAF* smaf = NULL;
	static int refrate = 10;

	static int inputtingnum;
	static char inputtingtype;
	static BOOL isinputting = FALSE;
	static BOOL isinputfirst;

	static int sndy;
	static int sndch;

	static int pch = 0;

	switch(msg)
	{
	case WM_CREATE:
		{
			TCHAR fn[MAX_PATH];

			GetModuleFileName(NULL, fn, sizeof(fn));
			lstrcpy((LPTSTR)basename(fn), "DefMA3_16.vm3");

			LoadVoices(fn);
			InitMidiMode();
		}

		break;
	case WM_MOUSEWHEEL:
		if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
		{
			if((short)HIWORD(wParam) > 0)
				PostMessage(hWnd, WM_HSCROLL, SB_PAGELEFT, 0);
			else
				PostMessage(hWnd, WM_HSCROLL, SB_PAGERIGHT, 0);
		}else
		{
			if((short)HIWORD(wParam) > 0)
				PostMessage(hWnd, WM_VSCROLL, SB_PAGEUP, 0);
			else
				PostMessage(hWnd, WM_VSCROLL, SB_PAGEDOWN, 0);
		}
		break;
	case WM_KEYDOWN:
		if((int)wParam == VK_SHIFT)
			if(!isinputting)
			{
				isinputting = TRUE;
				isinputfirst = TRUE;
				inputtingnum = 0;
				inputtingtype = 0;
				InvalidateRect(hWnd, NULL, FALSE);
			}

		if(isinputting)
		{
			if('0' <= (int)wParam && (int)wParam <= '9')
			{
				if(isinputfirst)
				{
					isinputfirst = FALSE;
					inputtingnum = 0;
				}
				inputtingnum = inputtingnum * 10 + (int)wParam - '0';
				InvalidateRect(hWnd, NULL, FALSE);
			}
			if((int)wParam == VK_OEM_MINUS)
			{
				inputtingnum = -inputtingnum;
				InvalidateRect(hWnd, NULL, FALSE);
			}
			if((int)wParam == VK_BACK || (int)wParam == VK_DELETE)
			{
				inputtingnum = inputtingnum / 10;
				InvalidateRect(hWnd, NULL, FALSE);
			}
			if(('A' <= (int)wParam && (int)wParam <= 'Z') || (int)wParam == VK_ESCAPE)
			{
				inputtingtype = (char)wParam;
				isinputfirst = TRUE;
				switch(inputtingtype)
				{
				case 'R':
					inputtingnum = refrate;
					break;
				case 'D':
					inputtingnum = DrawFunc;
					break;
				case 'H':
					inputtingnum = RollHeight;
					break;
				case 'T':
					inputtingnum = EmuGetTempo();
					break;
				case 'K':
					inputtingnum = EmuGetKeypitch();
					break;
				case 'W':
					inputtingnum = RollWidth;
					break;
				case 'F':
					inputtingnum = FollowPlay;
					break;
				case 'S':
					inputtingnum = -FollowSpace;
					break;
				case 'G':
					if(smaf)
						inputtingnum = PlayingPos;
					break;
				case 'V':
					inputtingnum = EmuGetVolume();
					break;
				case 'M':
					if(smaf)
						inputtingnum = GetScrollPos(hWnd, SB_HORZ) * RollWidth;
					break;
				case 'O':
					inputtingnum = PlayVisibleOnly;
					break;
				}
				InvalidateRect(hWnd, NULL, FALSE);
			}
		}else
		{
			if(!(GetAsyncKeyState(VK_SHIFT) & 0x8000) && !(GetAsyncKeyState(VK_CONTROL) & 0x8000) && !(GetAsyncKeyState(VK_MENU) & 0x8000))
			{
				switch((int)wParam)
				{
				case '0':	VisibleTrack = -1;	InvalidateRect(hWnd, NULL, FALSE);	break;

				case '1':	VisibleTrack = 0;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '2':	VisibleTrack = 1;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '3':	VisibleTrack = 2;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '4':	VisibleTrack = 3;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '5':	VisibleTrack = 4;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '6':	VisibleTrack = 5;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '7':	VisibleTrack = 6;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case '8':	VisibleTrack = 7;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'Q':	VisibleTrack = 8;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'W':	VisibleTrack = 9;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'E':	VisibleTrack = 10;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'R':	VisibleTrack = 11;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'T':	VisibleTrack = 12;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'Y':	VisibleTrack = 13;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'U':	VisibleTrack = 14;	InvalidateRect(hWnd, NULL, FALSE);	break;
				case 'I':	VisibleTrack = 15;	InvalidateRect(hWnd, NULL, FALSE);	break;
				}
				if(PlayVisibleOnly)
					EmuMIDISetSoloCh(VisibleTrack);
			}
		}

		switch((int)wParam)
		{
		case VK_LEFT:
			if(smaf)
			{
				UINT movval = 1000;

				if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
					movval = 100;

				if((smaf->start->time - smaf->start->duration) + movval <= PlayingPos)
					seek(smaf, PlayingPos - movval, emusounding || midisounding, FALSE);
				else
					seek(smaf, smaf->start->time - smaf->start->duration, emusounding || midisounding, FALSE);

				{
					TCHAR ts[32];

					wsprintf(ts, "MMF Tool %06ums", PlayingPos);
					SetWindowText(hWnd, ts);
				}
			}
			break;
		case VK_RIGHT:
			if(smaf)
			{
				UINT movval = 1000;

				if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
					movval = 100;

				seek(smaf, PlayingPos + movval, emusounding || midisounding, FALSE);

				{
					TCHAR ts[32];

					wsprintf(ts, "MMF Tool %06ums", PlayingPos);
					SetWindowText(hWnd, ts);
				}
			}
			break;
		}
		break;
	case WM_KEYUP:
		if((int)wParam == VK_SHIFT)
			if(isinputting)
			{
				isinputting = FALSE;
				switch(inputtingtype)
				{
				case 'R':
					if(0 < inputtingnum)
						refrate = inputtingnum;
					if(emusounding || midisounding)
					{
						KillTimer(hWnd, 0);
						SetTimer(hWnd, 0, refrate, NULL);
					}
					break;
				case 'D':
					if(0 <= inputtingnum && inputtingnum <= 2)
						DrawFunc = inputtingnum;
					break;
				case 'H':
					if(0 < inputtingnum)
						RollHeight = inputtingnum;
					break;
				case 'T':
					if(70 <= inputtingnum && inputtingnum <= 130)
					{
						// e|(%) between 70 to 130.
						EmuSetTempo(inputtingnum);
					}
					break;
				case 'K':
					if(-12 <= inputtingnum && inputtingnum <= 12)
					{
						EmuSetKeypitch(inputtingnum);
					}
					break;
				case 'W':
					if(0 < inputtingnum)
					{
						UINT curr;

						curr = GetScrollPos(hWnd, SB_HORZ) * RollWidth;

						RollWidth = inputtingnum;
						if(smaf)
							SetScrollPos(hWnd, SB_HORZ, (int)curr / (int)RollWidth, TRUE);
					}
					break;
				case 'F':
					if(0 <= inputtingnum && inputtingnum <= 2)
						FollowPlay = inputtingnum;
					break;
				case 'S':
					FollowSpace = -inputtingnum;
					break;
				case 'G':
					if(smaf && 0 <= inputtingnum)
					{
						seek(smaf, inputtingnum, emusounding || midisounding, TRUE);

						{
							TCHAR ts[32];

							wsprintf(ts, "MMF Tool %06ums", PlayingPos);
							SetWindowText(hWnd, ts);
						}
					}
					break;
				case 'V':
					if(0 <= inputtingnum && inputtingnum <= 127)
					{
						//  between 0 to 127
						EmuSetVolume(inputtingnum);
					}
					break;
				case 'M':
					if(smaf && 0 <= inputtingnum)
						SetScrollPos(hWnd, SB_HORZ, (int)(inputtingnum) / (int)RollWidth, TRUE);
					break;
				case 'O':
					if(0 <= inputtingnum && inputtingnum <= 1)
					{
						PlayVisibleOnly = inputtingnum;
						if(PlayVisibleOnly)
							EmuMIDISetSoloCh(VisibleTrack);
						else
							EmuMIDISetSoloCh(-1);
					}
					break;
				}
				InvalidateRect(hWnd, NULL, FALSE);
			}
		break;
	case WM_HSCROLL:
	case WM_VSCROLL:
		{
			SCROLLINFO scr;

			scr.cbSize = sizeof(scr);
			scr.fMask = SIF_ALL;
			if(msg == WM_HSCROLL)
				GetScrollInfo(hWnd, SB_HORZ, &scr);
			if(msg == WM_VSCROLL)
				GetScrollInfo(hWnd, SB_VERT, &scr);

			switch(LOWORD(wParam))
			{
			//case SB_TOP:
			case SB_LEFT:
				scr.nPos = scr.nMin;
				break;
			//case SB_BOTTOM:
			case SB_RIGHT:
				scr.nPos = scr.nMax;
				break;
			//case SB_LINEUP:
			case SB_LINELEFT:
				if(scr.nMin < scr.nPos)
					scr.nPos--;
				break;
			//case SB_LINEDOWN:
			case SB_LINERIGHT:
				if(scr.nPos < scr.nMax - 1)
					scr.nPos++;
				break;
			//case SB_PAGEUP:
			case SB_PAGELEFT:
				scr.nPos -= scr.nPage;
				break;
			//case SB_PAGEDOWN:
			case SB_PAGERIGHT:
				scr.nPos += scr.nPage;
				break;
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK:
				scr.nPos = HIWORD(wParam);
				break;
			}

			scr.fMask = SIF_POS;
			if(msg == WM_HSCROLL)
				SetScrollInfo(hWnd, SB_HORZ, &scr, TRUE);
			if(msg == WM_VSCROLL)
				SetScrollInfo(hWnd, SB_VERT, &scr, TRUE);

			InvalidateRect(hWnd, NULL, FALSE);
		}

		return 0;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;
			RECT rc;

			HDC hdcbg;
			HBITMAP oldbmp;
			BITMAPINFO bi = {0};

			hdc = BeginPaint(hWnd, &ps);
			GetClientRect(hWnd, &rc);

			switch(DrawFunc)
			{
			case 0:
				break;
			case 1:
				hdcbg = hdc;
				hdc = CreateCompatibleDC(hdcbg);

				bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				bi.bmiHeader.biPlanes = 1;
				bi.bmiHeader.biWidth = rc.right;
				bi.bmiHeader.biHeight = rc.bottom;
				bi.bmiHeader.biBitCount = 24;
				oldbmp = SelectObject(hdc, CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, NULL, NULL, 0));

				break;
			case 2:
				hdcbg = hdc;
				hdc = CreateCompatibleDC(hdcbg);

				oldbmp = SelectObject(hdc, CreateCompatibleBitmap(hdcbg, rc.right, rc.bottom));

				break;
			}

			draw(hWnd, hdc, rc.right, rc.bottom, smaf, PlayingPos);
			if(isinputting)
			{
				MoveToEx(hdc, 0, 0, NULL);
				SaveDC(hdc);
				SetBkMode(hdc, OPAQUE);
				SetBkColor(hdc, RGB(255, 255, 255));
				SetTextColor(hdc, RGB(0, 0, 0));
				switch(inputtingtype)
				{
				case 0:
					break;
				case VK_ESCAPE:
					PutText(hdc, "(Escape)");
					break;
				case 'R':	PutText(hdc, "Refresh Rate: %d", inputtingnum);					break;
				case 'D':	PutText(hdc, "Draw Function: %d", inputtingnum);				break;
				case 'H':	PutText(hdc, "PianoRoll's Height: %d", inputtingnum);			break;
				case 'T':	PutText(hdc, "Tempo: %d", inputtingnum);						break;
				case 'K':	PutText(hdc, "KeyPitch: %d", inputtingnum);						break;
				case 'W':	PutText(hdc, "PianoRoll's Width divider: %d", inputtingnum);	break;
				case 'F':	PutText(hdc, "Follow play: %d", inputtingnum);					break;
				case 'S':	PutText(hdc, "Follow space: %d", -inputtingnum);				break;
				case 'G':
					if(smaf)
						PutText(hdc, "GoTo: %d", inputtingnum);
					else
						PutText(hdc, "SMAF not opened", inputtingnum);
					break;
				case 'V':	PutText(hdc, "Volume: %d", inputtingnum);						break;
				case 'M':	PutText(hdc, "MoveTo: %d", inputtingnum);						break;
				case 'O':	PutText(hdc, "PlayVisibleOnly: %d", inputtingnum);				break;
				default:	PutText(hdc, "INVALID TYPE");									break;
				}
				RestoreDC(hdc, -1);
			}

			switch(DrawFunc)
			{
			case 0:
				break;
			case 1:
			case 2:
				BitBlt(hdcbg, 0, 0, rc.right, rc.bottom, hdc, 0, 0, SRCCOPY);

				DeleteObject(SelectObject(hdc, oldbmp));
				DeleteDC(hdc);

				break;
			}

			EndPaint(hWnd, &ps);
		}
		break;
	case WM_INITMENUPOPUP:
		{
			HMENU hm = (HMENU)wParam;
			int i;
			int max = GetMenuItemCount(hm);

			for(i = 0; i < max; i++)
			{
				MENUITEMINFO mii = {0};

				mii.cbSize = sizeof(mii);
				mii.fMask = MIIM_ID;
				GetMenuItemInfo(hm, i, TRUE, &mii);
				switch(mii.wID)
				{
				case ID_VTRACK1:
				case ID_VTRACK2:
				case ID_VTRACK3:
				case ID_VTRACK4:
				case ID_VTRACK5:
				case ID_VTRACK6:
				case ID_VTRACK7:
				case ID_VTRACK8:
				case ID_VTRACK9:
				case ID_VTRACK10:
				case ID_VTRACK11:
				case ID_VTRACK12:
				case ID_VTRACK13:
				case ID_VTRACK14:
				case ID_VTRACK15:
				case ID_VTRACK16:
					mii.fMask = MIIM_ID | MIIM_TYPE;
					mii.fType = MFT_OWNERDRAW;
					SetMenuItemInfo(hm, i, TRUE, &mii);
					break;
				}
			}
		}
		break;
	case WM_MEASUREITEM:
		{
			LPMEASUREITEMSTRUCT mis = (LPMEASUREITEMSTRUCT)lParam;

			if(mis->CtlType == ODT_MENU)
			{	// ͂Ђǂ
				HDC hmwdc;
				HDC hdc;
				NONCLIENTMETRICS ncm;
				HFONT hfold;
				SIZE s;
				TCHAR str[256];

				hmwdc = GetDC(hWnd);
				hdc = CreateCompatibleDC(hmwdc);

				ncm.cbSize = sizeof(ncm);
				SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(ncm), &ncm, 0);

				hfold = SelectObject(hdc, CreateFontIndirect(&ncm.lfMenuFont));

				getTrackMenuItem(mis->itemID, str, NULL, NULL);
				GetTextExtentPoint32(hdc, str, lstrlen(str), &s);
				mis->itemWidth = s.cx + 4;
				mis->itemHeight = s.cy + 4;

				DeleteObject(SelectObject(hdc, hfold));
				DeleteDC(hdc);
				ReleaseDC(hWnd, hmwdc);
			}
		}
		break;
	case WM_DRAWITEM:
		{
			LPDRAWITEMSTRUCT dis = (LPDRAWITEMSTRUCT)lParam;

			if(dis->CtlType == ODT_MENU)
			{
				TCHAR str[256];
				MENUITEMINFO mii = {0};
				COLORREF bg = 0xFF000000, fg;

				getTrackMenuItem(dis->itemID, str, &fg, &bg);

				if(bg != 0xFF000000)
				{
					SIZE s;

					SaveDC(dis->hDC);
					SetBkMode(dis->hDC, TRANSPARENT);

					if(dis->itemState & ODS_SELECTED)
					{
						SetTextColor(dis->hDC, fg);
						FillRectangle(dis->hDC, dis->rcItem.left, dis->rcItem.top, dis->rcItem.right, dis->rcItem.bottom, RGB(128, 0, 128));
						FillRectangle(dis->hDC, dis->rcItem.left + 1, dis->rcItem.top + 1, dis->rcItem.right - 1, dis->rcItem.bottom - 1, bg);
					}else
						FillRectangle(dis->hDC, dis->rcItem.left, dis->rcItem.top, dis->rcItem.right, dis->rcItem.bottom, GetSysColor(COLOR_3DFACE));

					GetTextExtentPoint32(dis->hDC, str, lstrlen(str), &s);

					TextOut(dis->hDC, dis->rcItem.left + 2, dis->rcItem.top + (dis->rcItem.bottom - dis->rcItem.top - s.cy) / 2, str, lstrlen(str));

					RestoreDC(dis->hDC, -1);
				}
			}
		}
		break;
	case WM_DROPFILES:
		{
			HDROP hd = (HDROP)wParam;
			TCHAR fn[MAX_PATH];

			DragQueryFile(hd, 0, fn, sizeof(fn));
			DragFinish(hd);

			OpenSMAFCommand(hWnd, &smaf, fn);
		}
		break;
	case WM_COMMAND:
		switch(HIWORD(wParam))
		{
		case 0:	// Command
		case 1:	// Accelerator
			switch(LOWORD(wParam))
			{
			case ID_OPENSMAF:
				{
					OPENFILENAME ofn = {0};
					TCHAR fname[MAX_PATH];

					ofn.lStructSize = sizeof(ofn);
					ofn.Flags = OFN_FILEMUSTEXIST;
					ofn.hInstance = GetModuleHandle(NULL);
					ofn.hwndOwner = hWnd;
					ofn.lpstrTitle = "Open SMAF";
					fname[0] = '\0';
					ofn.lpstrFile = fname;
					ofn.nMaxFile = sizeof(fname);
					ofn.lpstrFilter = "SMAF file(*.mmf)\0*.mmf\0";

					if(GetOpenFileName(&ofn))
						OpenSMAFCommand(hWnd, &smaf, fname);
				}
				break;
			case ID_CLOSE:
				if(smaf)
				{
					if(emusounding)
						SendMessage(hWnd, WM_COMMAND, ID_STOP, 0);
					if(midisounding)
						SendMessage(hWnd, WM_COMMAND, ID_MIDISTOP, 0);

					smaf = FreeSmaf(smaf);

					SendMessage(hDetailWnd, WM_USER, 0, (LPARAM)smaf);

					SetWindowText(hWnd, "MMF Tool");

					InvalidateRect(hWnd, NULL, FALSE);
				}
				break;
			case ID_EXIT:
				PostMessage(hWnd, WM_CLOSE, 0, 0);
				break;
			case ID_HELPINDEX:
				MessageBox(hWnd, "There are no help.", "sorry", MB_ICONERROR | MB_OK);
				break;
			case ID_ABOUT:
				DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT), hWnd, MessageDlgWndProc, 0);
				break;
			case ID_PLAY:
				if(smaf && !emusounding && !midisounding)
				{
					UINT startfrom = PlayingPos;

					SetWindowText(hWnd, "MMF Tool <Starting play...>");

					if(startfrom < smaf->start->time - smaf->start->duration)
						startfrom = smaf->start->time - smaf->start->duration;
					EmuSMAFPlay(smaf, startfrom - (smaf->start->time - smaf->start->duration));
					emusounding = TRUE;
					SetTimer(hWnd, 0, refrate, NULL);
				}
				break;
			case ID_STOP:
				if(smaf && emusounding)
				{
					SetWindowText(hWnd, "MMF Tool <Stopping play...>");

					EmuSMAFStop();
					InitMidiMode();
					emusounding = FALSE;
					KillTimer(hWnd, 0);

					{
						TCHAR ts[32];

						wsprintf(ts, "MMF Tool %06ums", PlayingPos);
						SetWindowText(hWnd, ts);
					}

					InvalidateRect(hWnd, NULL, FALSE);
				}
				break;
			case ID_REWIND:
				if(smaf)
				{
					seek(smaf, smaf->start->time - smaf->start->duration, emusounding || midisounding, FALSE);

					{
						TCHAR ts[32];

						wsprintf(ts, "MMF Tool %06ums", PlayingPos);
						SetWindowText(hWnd, ts);
					}
				}
				break;
			case ID_DETAILWND:
				ShowWindow(hDetailWnd, SW_SHOW);
				SetForegroundWindow(hDetailWnd);
				break;
			case ID_VTRACKALL:	VisibleTrack = -1;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK1:	VisibleTrack = 0;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK2:	VisibleTrack = 1;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK3:	VisibleTrack = 2;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK4:	VisibleTrack = 3;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK5:	VisibleTrack = 4;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK6:	VisibleTrack = 5;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK7:	VisibleTrack = 6;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK8:	VisibleTrack = 7;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK9:	VisibleTrack = 8;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK10:	VisibleTrack = 9;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK11:	VisibleTrack = 10;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK12:	VisibleTrack = 11;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK13:	VisibleTrack = 12;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK14:	VisibleTrack = 13;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK15:	VisibleTrack = 14;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_VTRACK16:	VisibleTrack = 15;	InvalidateRect(hWnd, NULL, FALSE);	break;
			case ID_MIDIPLAY:
				if(smaf && !midisounding && !emusounding)
				{
					UINT startfrom = PlayingPos;

					SetWindowText(hWnd, "MMF Tool <Starting play...>");

					if(startfrom < smaf->start->time - smaf->start->duration)
						startfrom = smaf->start->time - smaf->start->duration;
					EmuMIDIPlay(smaf, startfrom - (smaf->start->time - smaf->start->duration));
					midisounding = TRUE;
					SetTimer(hWnd, 0, refrate, NULL);
					SetTimer(hWnd, 1, 10, NULL);
				}
				break;
			case ID_MIDISTOP:
				if(smaf && midisounding)
				{
					SetWindowText(hWnd, "MMF Tool <Stopping play...>");

					EmuMIDIStop();
					midisounding = FALSE;
					KillTimer(hWnd, 0);
					KillTimer(hWnd, 1);

					SetWindowText(hWnd, "MMF Tool");

					InvalidateRect(hWnd, NULL, FALSE);
				}
				break;
			case ID_MIDIRESET:
				InitMidiMode();

				break;
			}
			break;
		}
		break;
	case WM_LBUTTONDOWN:
		if(smaf)
		{
			UINT pos = (GetScrollPos(hWnd, SB_HORZ) + LOWORD(lParam)) * RollWidth;
			if(pos < smaf->start->time - smaf->start->duration)
				pos = smaf->start->time - smaf->start->duration;
			seek(smaf, pos, emusounding || midisounding, FALSE);

			{
				TCHAR ts[32];

				wsprintf(ts, "MMF Tool %06ums", PlayingPos);
				SetWindowText(hWnd, ts);
			}
		}
		break;
	case WM_MBUTTONDOWN:
		{
			HMENU hm[9] = {0};
			POINT pt;
			int gr;
			int ret;

			hm[0] = CreatePopupMenu();

			for(gr = 0; gr < 8; gr++)
			{
				int i;

				hm[gr + 1] = CreatePopupMenu();

				for(i = 0; i < 16; i++)
				{
					char ts[32];

					wsprintf(ts, "%d: %s", gr * 16 + i, voices[gr * 16 + i].name);
					InsertMenu(hm[gr + 1], i, MF_BYPOSITION | MF_STRING | ((pch == gr * 16 + i) ? MF_CHECKED : 0), gr * 16 + i + 100, ts);
				}

				{
					char ts[32];
					wsprintf(ts, "Inst# %d - %d", gr * 16, gr * 16 + 15);
					InsertMenu(hm[0], gr, MF_BYPOSITION | MF_POPUP | ((gr * 16 <= pch && pch <= gr * 16 + 15) ? MF_CHECKED : 0), (UINT)hm[gr + 1], ts);
				}
			}

			GetCursorPos(&pt);
			ret = (int)TrackPopupMenu(hm[0], TPM_RETURNCMD, pt.x, pt.y, 0, hWnd, NULL);

			for(gr = 0; gr < 8; gr++)
				DestroyMenu(hm[gr + 1]);

			DestroyMenu(hm[0]);

			if(ret >= 100)
				pch = ret - 100;
		}
		break;
	case WM_RBUTTONDOWN:
		{
			int y = (GetScrollPos(hWnd, SB_VERT) + HIWORD(lParam)) / RollHeight;

			// F#-6 made.

			if(emusounding)
				break;

			if(128 <= y)
				break;

			y = 127 - y;

			{
				HDC hdc;

				hdc = GetDC(hWnd);
				SaveDC(hdc);
				SetBkMode(hdc, OPAQUE);
				SetBkColor(hdc, RGB(0,0,0));
				SetTextColor(hdc, RGB(255,255,255));
				MoveToEx(hdc, 0, 0, NULL);
				PutText(hdc, "%03d", y);
				RestoreDC(hdc, -1);
				ReleaseDC(hWnd, hdc);
			}

			// ?
			/*
			{
				BYTE cmd[] = {0xF0, 0x41, 0x79, 0x06, 0x7F, 0x07, 0x00, 0xF7};
				if(EmuSetMidiMsg(cmd, sizeof(cmd)))
				{
					MessageBox(hWnd, "SetMidiMsgXn failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
			*/

			// MasterVolume
			EmuMIDISetVolume();

			// ready.

			if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
			{
				UINT tr = (VisibleTrack == -1) ? 0 : VisibleTrack;

				// CC BankSelect
				{
					BYTE cmd[] = {0xB0 | tr, 0x00, 124};
					if(GetAsyncKeyState(VK_SHIFT) & 0x8000)
						cmd[2] = 125;
					if(EmuSetMidiMsg(cmd, 3))
					{
						MessageBox(hWnd, "SetMidiMsg5 failed", "Error", MB_ICONERROR | MB_OK);
						break;
					}
				}
				{
					BYTE cmd[] = {0xB0 | tr, 0x20, 0};
					if(EmuSetMidiMsg(cmd, 3))
					{
						MessageBox(hWnd, "SetMidiMsg6 failed", "Error", MB_ICONERROR | MB_OK);
						break;
					}
				}
			}
			// PC
			if(VisibleTrack == -1)
			{
				BYTE cmd[] = {0xC0, pch};
				if(EmuSetMidiMsg(cmd, 2))
				{
					MessageBox(hWnd, "SetMidiMsg7 failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
			// NoteON
			{
				BYTE cmd[] = {0x90, y, 0x7F};
				sndch = 0;
				if(VisibleTrack != -1)
				{
					cmd[0] = 0x90 | VisibleTrack;
					sndch = VisibleTrack;
				}
				if(EmuSetMidiMsg(cmd, 3))
				{
					MessageBox(hWnd, "SetMidiMsg8[sounding note] failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}

			sndy = y;
		}
		break;
	case WM_RBUTTONUP:
		{
			if(emusounding)
				break;

			if(sndy == -1)
				break;

			/*
			// CC BankSelect
			{
				BYTE cmd[] = {0xB0, 0x00, 0x7C};
				if(EmuSetMidiMsg(cmd, sizeof(cmd)))
				{
					MessageBox(hWnd, "SetMidiMsg5 failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
			{
				BYTE cmd[] = {0xB0, 0x20, 0x00};
				if(EmuSetMidiMsg(cmd, sizeof(cmd)))
				{
					MessageBox(hWnd, "SetMidiMsg6 failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
			// PC
			{
				BYTE cmd[] = {0xC0, 0};
				if(EmuSetMidiMsg(cmd, sizeof(cmd)))
				{
					MessageBox(hWnd, "SetMidiMsg7 failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
			*/
			// NoteOFF
			{
				BYTE cmd[] = {0x90 | sndch, sndy, 0x00};
				if(EmuSetMidiMsg(cmd, sizeof(cmd)))
				{
					MessageBox(hWnd, "SetMidiMsg8[sounding] failed", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
		}
		break;
	case WM_TIMER:
		if(emusounding)
		{
			static DWORD lastrefresh = 0;

			PlayingPos = EmuGetCurrentPos() + (smaf->start->time - smaf->start->duration);

			{
				TCHAR ts[256];
				DWORD emuinfo = EmuInfo();

				wsprintf(ts, "MMF Tool EMU:%06ums ", PlayingPos);
				if(emuinfo & 0x01)
					lstrcat(ts, "LED ");
				if(emuinfo & 0x02)
					lstrcat(ts, "Vib ");
				SetWindowText(hWnd, ts);
			}

			switch(EmuStatus())
			{
			case EMUSTATUS_PLAYING:
				break;
			case EMUSTATUS_READY:	// music stop.
				SendMessage(hWnd, WM_COMMAND, ID_STOP, 0);
				break;
			default:	// z͈̔͊OłB
				SendMessage(hWnd, WM_COMMAND, ID_STOP, 0);
				MessageBox(hWnd, "Status is NOT 4(Playing) or 3(Ready)\nSTOPPED.", "Error", MB_ICONERROR | MB_OK);
				break;
			}

			if(FollowPlay == 1 || (FollowPlay == 2 && GetAsyncKeyState(VK_SHIFT) & 0x8000))
			{
				SetScrollPos(hWnd, SB_HORZ, (int)(PlayingPos + FollowSpace) / (int)RollWidth, TRUE);
			}
			InvalidateRect(hWnd, NULL, FALSE);

			if(lastrefresh + 100 < GetTickCount())
			{
				lastrefresh = GetTickCount();
				InvalidateRect(hDetailWnd, NULL, FALSE);
			}
		}

		if(midisounding)
		{
			if(wParam == 0)	// ID:0 = Redraw
			{
				static DWORD lastrefresh = 0;

				{
					TCHAR ts[256];
					DWORD emuinfo = EmuInfo();

					wsprintf(ts, "MMF Tool MIDI:%06ums", PlayingPos);
					SetWindowText(hWnd, ts);
				}

				if(FollowPlay == 1 || (FollowPlay == 2 && GetAsyncKeyState(VK_SHIFT) & 0x8000))
				{
					SetScrollPos(hWnd, SB_HORZ, (int)(PlayingPos + FollowSpace) / (int)RollWidth, TRUE);
				}
				InvalidateRect(hWnd, NULL, FALSE);

				if(lastrefresh + 100 < GetTickCount())
				{
					lastrefresh = GetTickCount();
					InvalidateRect(hDetailWnd, NULL, FALSE);
				}
			}
			if(wParam == 1)	// ID:1 = MIDIProc
			{
				EmuMIDIProc();

				PlayingPos = EmuGetCurrentPos();

				switch(EmuStatus())
				{
				case EMUSTATUS_PLAYING:
					break;
				case EMUSTATUS_READY:	// music stop.
					SendMessage(hWnd, WM_COMMAND, ID_MIDISTOP, 0);
					break;
				default:	// z͈̔͊OłB
					SendMessage(hWnd, WM_COMMAND, ID_MIDISTOP, 0);
					MessageBox(hWnd, "Status is NOT 4(Playing) or 3(Ready)\nSTOPPED.", "Error", MB_ICONERROR | MB_OK);
					break;
				}
			}
		}

		break;
	case WM_SIZE:
		{
			SCROLLINFO si;

			si.cbSize = sizeof(si);
			si.fMask = SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
			si.nMin = 0;
			si.nMax = 65535;
			si.nPage = 10;
			SetScrollInfo(hWnd, SB_HORZ, &si, TRUE);

			si.nMin = 0;
			si.nMax = 256 * 8;
			si.nPage = 10;
			SetScrollInfo(hWnd, SB_VERT, &si, TRUE);

			InvalidateRect(hWnd, NULL, FALSE);
		}
		break;
	case WM_CLOSE:
		if(smaf)
			SendMessage(hWnd, WM_COMMAND, ID_CLOSE, 0);
		break;
	case WM_DESTROY:
		PostQuitMessage(0);
		DestroyWindow(hDetailWnd);
		if(voices)
			hfree(voices);
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

BOOL initWnd(void)
{
	WNDCLASSEX wc = {0};
	
	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.lpfnWndProc = MainWndProc;
	wc.lpszClassName = "MMFToolWC";
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN);

	if(!RegisterClassEx(&wc))
		return FALSE;

	// I don't like WS_OVERLAPPED.
	if(!(hMainWnd = CreateWindowEx(WS_EX_CLIENTEDGE | WS_EX_ACCEPTFILES, wc.lpszClassName, "MMF Tool", WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_SIZEBOX | WS_MAXIMIZEBOX | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, wc.hInstance, NULL)))
		return FALSE;

	wc.cbSize = sizeof(wc);
	wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.lpfnWndProc = DetailWndProc;
	wc.lpszClassName = "MMFToolDetailWC";
	wc.lpszMenuName = NULL;

	if(!RegisterClassEx(&wc))
		return FALSE;

	if(!(hDetailWnd = CreateWindowEx(WS_EX_CLIENTEDGE, wc.lpszClassName, "MMF Tool - DetailView", WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_SIZEBOX | WS_MAXIMIZEBOX | WS_HSCROLL | WS_VSCROLL, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, wc.hInstance, NULL)))
		return FALSE;

	return TRUE;
}

void WinMainCRTStartup(void)
{
	if(!EmuInitialize())
		ExitProcess(0);

	if(!initWnd())
		ExitProcess(0);

	{
		MSG msg;
		HACCEL at;

		at = LoadAccelerators(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MAIN));

		while(GetMessage(&msg, NULL, 0, 0))
		{
			if(!TranslateAccelerator(msg.hwnd, at, &msg))
				TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		EmuUninitialize();

		ExitProcess(msg.wParam);
	}
}
